from datetime import datetime, timedelta

from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Label, Static

from bagels.components.modules.welcome import Welcome
from bagels.components.modules.accountmode import AccountMode
from bagels.components.modules.datemode import DateMode
from bagels.components.modules.incomemode import IncomeMode
from bagels.components.modules.insights import Insights
from bagels.components.modules.records import Records
from bagels.components.modules.templates import Templates
from bagels.config import CONFIG
from bagels.managers.accounts import get_accounts_count, get_all_accounts
from bagels.managers.categories import get_categories_count
from bagels.utils.format import format_period_to_readable


class Home(Static):
    filter = {
        "offset": 0,
        "offset_type": CONFIG.defaults.period,
    }

    BINDINGS = [
        Binding("left", "dec_offset", "Previous", show=False),
        Binding("right", "inc_offset", "Next", show=False),
        Binding(
            CONFIG.hotkeys.home.cycle_offset_type, "cycle_offset_type", "", show=False
        ),
        Binding(
            CONFIG.hotkeys.home.toggle_income_mode,
            "toggle_income_mode",
            "Income/Expense",
            show=False,
        ),
        Binding(
            CONFIG.hotkeys.home.select_prev_account,
            "select_prev_account",
            "Select previous account",
            show=False,
        ),
        Binding(
            CONFIG.hotkeys.home.select_next_account,
            "select_next_account",
            "Select next account",
            show=False,
        ),
        Binding("1", "select_template_1", "Template 1", show=False),
        Binding("2", "select_template_2", "Template 2", show=False),
        Binding("3", "select_template_3", "Template 3", show=False),
        Binding("4", "select_template_4", "Template 4", show=False),
        Binding("5", "select_template_5", "Template 5", show=False),
        Binding("6", "select_template_6", "Template 6", show=False),
        Binding("7", "select_template_7", "Template 7", show=False),
        Binding("8", "select_template_8", "Template 8", show=False),
        Binding("9", "select_template_9", "Template 9", show=False),
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, id="home-page")
        self.isReady = get_accounts_count() and get_categories_count()
        accounts = get_all_accounts()
        self.mode = {
            "isIncome": False,
            "date": datetime.now(),
            "accountId": {
                "default_value": None,
                "default_value_text": "Select account",
            },
        }
        if accounts:
            self.mode["accountId"]["default_value"] = accounts[0].id
            self.mode["accountId"]["default_value_text"] = accounts[0].name
        self.accounts_indices = {"index": 0, "count": len(accounts)}
        self.accounts = accounts
        self.accounts_module = AccountMode(parent=self)
        self.date_mode_module = DateMode(parent=self)
        self.income_mode_module = IncomeMode(parent=self)
        self.record_module = Records(parent=self)
        self.insights_module = Insights(parent=self)
        self.templates_module = Templates(parent=self)

    def on_mount(self) -> None:
        self.app.watch(self.app, "layout", self.on_layout_change)
        if hasattr(self, "record_module"):
            self.record_module.focus()

    # -------------- Helpers ------------- #

    def rebuild(self) -> None:
        self.record_module.rebuild()
        self.insights_module.rebuild()
        self.accounts_module.rebuild()
        self.income_mode_module.rebuild()
        self.date_mode_module.rebuild()

    def get_filter_label(self) -> str:
        return format_period_to_readable(self.filter)

    def update_filter_label(self, label: Label) -> None:
        string = format_period_to_readable(self.filter)
        label.update(string)

    def get_target_date(self) -> datetime:
        """Returns the date that new records will be based on"""
        offset = self.filter["offset"]
        offset_type = self.filter["offset_type"]
        today = datetime.now()
        if offset == 0:
            return today

        match offset_type:
            case "year":
                return today
            case "month":
                # Get first day of current month
                first_of_month = today.replace(day=1)
                # Add months by adding days and resetting to first of month
                target_date = first_of_month + timedelta(days=32 * offset)
                return target_date.replace(day=1)
            case "week":
                # Get first day of current week
                days_to_first = (
                    today.weekday() - CONFIG.defaults.first_day_of_week
                ) % 7
                first_of_week = today - timedelta(days=days_to_first)
                # Add weeks
                return first_of_week + timedelta(weeks=offset)
            case "day":
                return today + timedelta(days=offset)
        return today

    def _update_date(self) -> None:
        self.mode["date"] = self.get_target_date()

    def set_target_date(self, date: datetime) -> None:
        self.mode["date"] = date

        # we also need to update the filters
        self.filter["offset_type"] = "day"
        offset = (date - datetime.now()).days
        self.filter["offset"] = offset + 1

        # rebuild
        self.rebuild()

    # region Callbacks
    # ------------- Callbacks ------------ #

    def on_layout_change(self, layout: str) -> None:
        layout_container = self.query(".home-modules-container")
        if len(layout_container) > 0:
            layout_container[0].set_classes(f"home-modules-container {layout}")

    def action_dec_offset(self) -> None:
        self.filter["offset"] -= 1
        self._update_date()
        self.rebuild()

    def action_inc_offset(self) -> None:
        if self.filter["offset"] < 0:
            self.filter["offset"] += 1
            self._update_date()
            self.rebuild()
        else:
            self.app.bell()

    def action_cycle_offset_type(self) -> None:
        # Define the cycle order
        cycle_order = ["day", "week", "month", "year"]

        # Get current index
        current_index = cycle_order.index(self.filter["offset_type"])

        next_index = (current_index + 1) % len(cycle_order)

        next_type = cycle_order[next_index]

        # calculate appropriate offset based on the target (mode date)
        match next_type:
            case "week":  # day -> week
                self.filter["offset"] = (
                    self.get_target_date() - datetime.now()
                ).days // 7 + 1
            case "month":  # week -> month
                self.filter["offset"] = (
                    self.get_target_date().year - datetime.now().year
                ) * 12 + (self.get_target_date().month - datetime.now().month)
            case _:
                self.filter["offset"] = 0

        self.filter["offset_type"] = next_type
        self._update_date()

        # Refresh table
        self.rebuild()

    def action_toggle_income_mode(self) -> None:
        self.mode["isIncome"] = not self.mode["isIncome"]
        self.income_mode_module.rebuild()
        self.insights_module.rebuild()

    def _select_account(self, dir: int) -> None:
        if self.accounts_indices["count"] > 0:
            new_index = (self.accounts_indices["index"] + dir) % self.accounts_indices[
                "count"
            ]
            self.accounts_indices["index"] = new_index
            self.mode["accountId"]["default_value"] = self.accounts[new_index].id
            self.mode["accountId"]["default_value_text"] = self.accounts[new_index].name
            self.accounts_module.rebuild()
            self.insights_module.rebuild()

    def action_select_prev_account(self) -> None:
        self._select_account(-1)

    def action_select_next_account(self) -> None:
        self._select_account(1)

    # region Templates
    # ------------- Template ------------- #

    def action_select_template_1(self) -> None:
        self.templates_module.select_template(1)

    def action_select_template_2(self) -> None:
        self.templates_module.select_template(2)

    def action_select_template_3(self) -> None:
        self.templates_module.select_template(3)

    def action_select_template_4(self) -> None:
        self.templates_module.select_template(4)

    def action_select_template_5(self) -> None:
        self.templates_module.select_template(5)

    def action_select_template_6(self) -> None:
        self.templates_module.select_template(6)

    def action_select_template_7(self) -> None:
        self.templates_module.select_template(7)

    def action_select_template_8(self) -> None:
        self.templates_module.select_template(8)

    def action_select_template_9(self) -> None:
        self.templates_module.select_template(9)

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        with Static(classes=f"home-modules-container v"):
            with Static(classes="left"):
                with Static(id="home-top-container"):
                    yield self.accounts_module
                    with Static(id="home-mode-container"):
                        yield self.income_mode_module
                        yield self.date_mode_module
                yield self.insights_module
            with Static(classes="right"):
                if self.isReady:
                    yield self.templates_module
                    yield self.record_module
                else:
                    yield Welcome()



from sqlalchemy.orm import sessionmaker, joinedload
from sqlalchemy import func, desc
from bagels.models.database.app import db_engine
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.managers.splits import create_split, get_splits_by_record_id, update_split
from bagels.managers.utils import get_start_end_of_period

Session = sessionmaker(bind=db_engine)


# region Create
def create_record(record_data: dict):
    session = Session()
    try:
        record = Record(**record_data)
        session.add(record)
        session.commit()
        session.refresh(record)
        session.expunge(record)
        return record
    finally:
        session.close()


def create_record_and_splits(record_data: dict, splits_data: list[dict]):
    session = Session()
    try:
        record = create_record(record_data)
        for split in splits_data:
            split["recordId"] = record.id
            create_split(split)
        return record
    finally:
        session.close()


# region Get
def get_record_by_id(record_id: int, populate_splits: bool = False):
    session = Session()
    try:
        query = session.query(Record).options(
            joinedload(Record.category), joinedload(Record.account)
        )

        if populate_splits:
            query = query.options(
                joinedload(Record.splits).options(
                    joinedload(Split.account), joinedload(Split.person)
                )
            )

        record = query.get(record_id)
        return record
    finally:
        session.close()


def get_record_total_split_amount(record_id: int):
    session = Session()
    try:
        splits = get_splits_by_record_id(record_id)
        return sum(split.amount for split in splits)
    finally:
        session.close()


def get_records(offset: int = 0, offset_type: str = "month"):
    session = Session()
    try:
        query = session.query(Record).options(
            joinedload(Record.category),
            joinedload(Record.account),
            joinedload(Record.transferToAccount),
            joinedload(Record.splits).options(
                joinedload(Split.account), joinedload(Split.person)
            ),
        )

        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)
        query = query.filter(
            Record.date >= start_of_period, Record.date < end_of_period
        )

        createdAt_column = getattr(Record, "createdAt")
        date_column = func.date(getattr(Record, "date"))
        query = query.order_by(date_column.desc(), createdAt_column.desc())

        records = query.all()
        return records
    finally:
        session.close()


def is_record_all_splits_paid(record_id: int):
    session = Session()
    try:
        splits = get_splits_by_record_id(record_id)
        return all(split.isPaid for split in splits)
    finally:
        session.close()


# region Update
def update_record(record_id: int, updated_data: dict):
    session = Session()
    try:
        record = session.query(Record).get(record_id)
        if record:
            for key, value in updated_data.items():
                setattr(record, key, value)
            session.commit()
            session.refresh(record)
            session.expunge(record)
        return record
    finally:
        session.close()


def update_record_and_splits(
    record_id: int, record_data: dict, splits_data: list[dict]
):
    session = Session()
    try:
        record = update_record(record_id, record_data)
        record_splits = get_splits_by_record_id(record_id)
        for index, split in enumerate(record_splits):
            update_split(split.id, splits_data[index])
        return record
    finally:
        session.close()


# region Delete
def delete_record(record_id: int):
    session = Session()
    try:
        record = session.query(Record).get(record_id)
        if record:
            session.delete(record)
            session.commit()
        return record
    finally:
        session.close()



from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Input, Label, Static, Switch

from bagels.components.autocomplete import AutoComplete, Dropdown, DropdownItem
from bagels.forms.form import Form, FormField


class Fields(Static):
    """Container for multiple form fields"""

    def __init__(self, fields: Form):  # should rename, but whatever
        super().__init__()
        self.form = fields

    def compose(self) -> ComposeResult:
        for field in self.form.fields:
            yield Field(field)


class Field(Static):
    """Individual form field that can be text, number, boolean, or autocomplete"""

    def __init__(self, field: FormField):
        super().__init__()
        self.field = field

        # Create base input widget
        self.input = Input(placeholder=field.placeholder or "", id=f"field-{field.key}")

        # Configure input based on field type
        match self.field.type:
            case "hidden":
                pass
            case "integer" | "number":
                self.input.type = self.field.type
                self.input.value = field.default_value or ""

            case "autocomplete":
                self.input.heldValue = field.default_value or ""
                self.input.value = field.default_value_text or field.default_value or ""

            case type_ if type_ != "boolean":
                self.input.value = field.default_value or ""

    def on_auto_complete_selected(self, event: AutoComplete.Selected) -> None:
        """Handle autocomplete selection"""
        self.screen.focus_next()

        # Find matching option and set held value
        for item in self.field.options.items:
            item_text = str(item.text or item.value)
            selected_dropdown_text = str(event.item.main)
            if item_text == selected_dropdown_text:
                self.input.heldValue = item.value
                break

    def compose(self) -> ComposeResult:
        if self.field.type == "hidden":
            # Hidden fields just need a static widget to hold the value
            self.input = Static(id=f"field-{self.field.key}")
            self.input.value = self.input.heldValue = self.field.default_value
            yield self.input
            return

        # Visible fields get a container with label and input
        with Container(classes="field-row", id=f"row-field-{self.field.key}"):
            yield Label(f"{self.field.title}", classes="label")

            if self.field.type == "autocomplete":
                # Build dropdown items list
                dropdown_items = [
                    DropdownItem(
                        item.text or item.value,
                        item.prefix or "",
                        item.postfix or "",
                    )
                    for item in self.field.options.items
                ]
                dropdown = Dropdown(
                    items=dropdown_items,
                    show_on_focus=True,
                    id=f"dropdown-{self.field.key}",
                    create_option=self.field.create_action,
                )

                yield AutoComplete(
                    self.input,
                    dropdown,
                    classes="field-autocomplete",
                    create_action=self.field.create_action,
                )

            elif self.field.type == "boolean":
                with Container(classes="switch-group"):
                    yield Label(str(self.field.labels[0]), classes="left")
                    yield Switch(
                        id=f"field-{self.field.key}",
                        value=self.field.default_value or False,
                    )
                    yield Label(str(self.field.labels[1]), classes="right")

            else:
                yield self.input



from textual import events
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import (
    Label,
)

from bagels.components.fields import Fields
from bagels.modals.base_widget import ModalContainer
from bagels.forms.form import Form
from bagels.utils.validation import validateForm


class InputModal(ModalScreen):
    def __init__(self, title: str, form: Form, *args, **kwargs):
        super().__init__(classes="modal-screen", *args, **kwargs)
        self.title = title
        self.form = form

    # --------------- Hooks -------------- #

    def on_key(self, event: events.Key):
        if event.key == "down":
            self.screen.focus_next()
        elif event.key == "up":
            self.screen.focus_previous()
        elif event.key == "enter":
            self.action_submit()
        elif event.key == "escape":
            self.dismiss(None)

    # ------------- Callbacks ------------ #

    def action_submit(self):
        resultForm, errors, isValid = validateForm(self, self.form)
        if isValid:
            self.dismiss(resultForm)
        else:
            previousErrors = self.query(".error")
            for error in previousErrors:
                error.remove()
            for key, value in errors.items():
                field = self.query_one(f"#row-field-{key}")
                field.mount(Label(value, classes="error"))

    # -------------- Compose ------------- #

    def compose(self) -> ComposeResult:
        yield ModalContainer(Fields(self.form))



import copy
from datetime import datetime

from rich.text import Text

from bagels.components.autocomplete import Dropdown
from bagels.managers.accounts import get_all_accounts_with_balance
from bagels.managers.categories import get_all_categories_by_freq
from bagels.managers.persons import create_person, get_all_persons
from bagels.managers.records import get_record_by_id, get_record_total_split_amount
from bagels.forms.form import Form, FormField, Option, Options


class RecordForm:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # ------------ Blueprints ------------ #

    FORM = Form(
        fields=[
            FormField(
                placeholder="Label",
                title="Label",
                key="label",
                type="string",
                is_required=True,
            ),
            FormField(
                title="Category",
                key="categoryId",
                type="autocomplete",
                options=Options(),
                is_required=True,
                placeholder="Select Category",
            ),
            FormField(
                placeholder="0.00",
                title="Amount",
                key="amount",
                type="number",
                min=0,
                is_required=True,
            ),
            FormField(
                title="Account",
                key="accountId",
                type="autocomplete",
                options=Options(),
                is_required=True,
                placeholder="Select Account",
            ),
            FormField(
                title="Type",
                key="isIncome",
                type="boolean",
                labels=["Expense", "Income"],
                default_value=False,
            ),
            FormField(
                placeholder="dd (mm) (yy)",
                title="Date",
                key="date",
                type="dateAutoDay",
                default_value=datetime.now().strftime("%d"),
            ),
        ]
    )

    SPLIT_FORM = Form(
        fields=[
            FormField(
                title="Person",
                key="personId",
                type="autocomplete",
                options=Options(),
                create_action=True,
                is_required=True,
                placeholder="Select Person",
            ),
            FormField(
                title="Amount",
                key="amount",
                type="number",
                min=0,
                is_required=True,
                placeholder="0.00",
            ),
            FormField(
                title="Paid",
                key="isPaid",
                type="hidden",
                default_value=False,
            ),
            FormField(
                title="Paid to account",
                key="accountId",
                type="hidden",
                options=Options(),
                placeholder="Select Account",
            ),
            FormField(
                title="Paid Date",
                key="paidDate",
                type="hidden",
                default_value=None,
            ),
        ]
    )

    # ----------------- - ---------------- #

    def __init__(self):
        self._populate_form_options()

    # -------------- Helpers ------------- #

    def _populate_form_options(self):
        accounts = get_all_accounts_with_balance()
        self.FORM.fields[3].options = Options(
            items=[
                Option(
                    text=account.name,
                    value=account.id,
                    postfix=Text(f"{account.balance}", style="yellow"),
                )
                for account in accounts
            ]
        )
        if accounts:
            self.FORM.fields[3].default_value = accounts[0].id
            self.FORM.fields[3].default_value_text = accounts[0].name

        categories = get_all_categories_by_freq()
        self.FORM.fields[1].options = Options(
            items=[
                Option(
                    text=category.name,
                    value=category.id,
                    prefix=Text("●", style=category.color),
                    postfix=(
                        Text(
                            (
                                f"↪ {category.parentCategory.name}"
                                if category.parentCategory
                                else ""
                            ),
                            style=category.parentCategory.color,
                        )
                        if category.parentCategory
                        else ""
                    ),
                )
                for category, _ in categories
            ]
        )
        people = get_all_persons()
        self.SPLIT_FORM.fields[0].options = Options(
            items=[Option(text=person.name, value=person.id) for person in people]
        )
        self.SPLIT_FORM.fields[3].options = Options(
            items=[Option(text=account.name, value=account.id) for account in accounts]
        )

    # ------------- Builders ------------- #

    def get_split_form(self, index: int, isPaid: bool = False) -> Form:
        split_form = copy.deepcopy(self.SPLIT_FORM)
        for field in split_form.fields:
            fieldKey = field.key
            field.key = f"{fieldKey}-{index}"
            if fieldKey == "isPaid":
                field.default_value = isPaid
            elif fieldKey == "accountId" and isPaid:
                field.type = "autocomplete"
            elif fieldKey == "paidDate" and isPaid:
                field.type = "dateAutoDay"
                field.default_value = datetime.now().strftime("%d")
        return split_form

    def get_filled_form(self, recordId: int) -> tuple[list, list]:
        """Return a copy of the form with values from the record"""
        filled_form = copy.deepcopy(self.FORM)
        record = get_record_by_id(recordId, populate_splits=True)

        for field in filled_form.fields:
            fieldKey = field.key
            value = getattr(record, fieldKey)

            match fieldKey:
                case "amount":
                    field.default_value = str(
                        round(value - get_record_total_split_amount(recordId), 2)
                    )
                case "date":
                    # if value is this month, simply set %d, else set %d %m %y
                    if value.month == datetime.now().month:
                        field.default_value = value.strftime("%d")
                    else:
                        field.default_value = value.strftime("%d %m %y")
                case "isIncome":
                    field.default_value = value
                case "categoryId":
                    field.default_value = record.category.id
                    field.default_value_text = record.category.name
                case "accountId":
                    field.default_value = record.account.id
                    field.default_value_text = record.account.name
                case _:
                    field.default_value = str(value) if value is not None else ""

        filled_splits = Form()
        for index, split in enumerate(record.splits):
            split_form = self.get_split_form(index, split.isPaid)
            for field in split_form.fields:
                fieldKey = field.key.split("-")[0]
                value = getattr(split, fieldKey)

                match fieldKey:
                    case "paidDate":
                        if value:
                            if value.month == datetime.now().month:
                                field.default_value = value.strftime("%d")
                            else:
                                field.default_value = value.strftime("%d %m %y")
                    case "accountId":
                        if split.account:
                            field.default_value = split.account.id
                            field.default_value_text = split.account.name
                    case "personId":
                        field.default_value = split.person.id
                        field.default_value_text = split.person.name
                    case "isPaid":
                        field.default_value = split.isPaid
                    case _:
                        field.default_value = str(value) if value is not None else ""

                filled_splits.fields.append(field)

        return filled_form, filled_splits

    def get_form(self, hidden_fields: dict = {}):
        """Return the base form with default values"""
        form = copy.deepcopy(self.FORM)
        for field in form.fields:
            key = field.key
            if key in hidden_fields:
                field.type = "hidden"
                if isinstance(hidden_fields[key], dict):
                    field.default_value = hidden_fields[key]["default_value"]
                    field.default_value_text = hidden_fields[key]["default_value_text"]
                else:
                    field.default_value = hidden_fields[key]
        return form



from datetime import datetime
from sqlalchemy import select
from sqlalchemy.orm import sessionmaker

from bagels.models.account import Account
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.models.database.app import db_engine

Session = sessionmaker(bind=db_engine)


def get_account_balance(accountId, session=None):
    """Returns the net balance of an account.

    Rules:
    - Consider all record "account" and split "account"
    - Records with isTransfer should consider both "account" and "transferToAccount"
    - Records and splits should be considered separately, unlike net figures which consider records and splits together.

    Args:
        accountId (int): The ID of the account to get the balance
        session (Session, optional): SQLAlchemy session to use. If None, creates a new session.
    """
    if session is None:
        session = Session()
        should_close = True
    else:
        should_close = False

    try:
        # Initialize balance
        balance = (
            session.query(Account).filter(Account.id == accountId).first().beginningBalance
        )

        # Get all records for this account
        records = session.query(Record).filter(Record.accountId == accountId).all()

        # Calculate balance from records
        for record in records:
            if record.isTransfer:
                # For transfers, subtract full amount (transfers out)
                balance -= record.amount
            elif record.isIncome:
                # For income records, add full amount
                balance += record.amount
            else:
                # For expense records, subtract full amount
                balance -= record.amount

        # Get all records where this account is the transfer destination
        transfer_to_records = (
            session.query(Record)
            .filter(Record.transferToAccountId == accountId, Record.isTransfer == True)
            .all()
        )

        # Add transfers into this account
        for record in transfer_to_records:
            balance += record.amount

        # Get all splits where this account is specified
        splits = session.query(Split).filter(Split.accountId == accountId).all()

        # Add paid splits (they represent money coming into this account)
        for split in splits:
            if split.isPaid:
                balance += split.amount

        return round(balance, 2)
    finally:
        if should_close:
            session.close()


def create_account(data):
    session = Session()
    try:
        new_account = Account(**data)
        session.add(new_account)
        session.commit()
        session.refresh(new_account)
        session.expunge(new_account)
        return new_account
    finally:
        session.close()


def _get_base_accounts_query(get_hidden=False):
    stmt = select(Account).filter(Account.deletedAt.is_(None))
    if not get_hidden:
        stmt = stmt.filter(Account.hidden.is_(False))
    else:
        stmt = stmt.order_by(Account.hidden)
    return stmt


def get_all_accounts(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        return session.scalars(stmt).all()
    finally:
        session.close()


def get_accounts_count(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        return len(session.scalars(stmt).all())
    finally:
        session.close()


def get_all_accounts_with_balance(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        accounts = session.scalars(stmt).all()
        for account in accounts:
            account.balance = get_account_balance(account.id, session)
        return accounts
    finally:
        session.close()


def get_account_balance_by_id(account_id):
    session = Session()
    try:
        return get_account_balance(account_id, session)
    finally:
        session.close()


def get_account_by_id(account_id):
    session = Session()
    try:
        return session.get(Account, account_id)
    finally:
        session.close()


def update_account(account_id, data):
    session = Session()
    try:
        account = session.get(Account, account_id)
        if account:
            for key, value in data.items():
                setattr(account, key, value)
            session.commit()
            session.refresh(account)
            session.expunge(account)
        return account
    finally:
        session.close()


def delete_account(account_id):
    session = Session()
    try:
        account = session.get(Account, account_id)
        if account:
            account.deletedAt = datetime.now()
            session.commit()
            return True
        return False
    finally:
        session.close()



from datetime import datetime
from rich.text import Text
from sqlalchemy import func, desc, select
from sqlalchemy.orm import joinedload, sessionmaker

from bagels.models.category import Category
from bagels.models.record import Record
from bagels.managers.utils import get_start_end_of_period
from bagels.models.database.app import db_engine

Session = sessionmaker(bind=db_engine)


def _get_base_categories_query(include_deleted=False):
    """Base query for categories with optional filtering of deleted entries."""
    stmt = select(Category)
    if not include_deleted:
        stmt = stmt.filter(Category.deletedAt.is_(None))
    return stmt


# region Get
def get_categories_count():
    """Count all categories excluding deleted ones."""
    session = Session()
    try:
        stmt = _get_base_categories_query()
        return len(session.scalars(stmt).all())
    finally:
        session.close()


def get_all_categories_tree():
    """Retrieve all categories in a hierarchical tree format."""
    session = Session()
    try:
        stmt = (
            _get_base_categories_query()
            .options(joinedload(Category.parentCategory))
            .order_by(Category.id)
        )
        categories = session.scalars(stmt).all()

        def build_category_tree(parent_id=None, depth=0):
            result = []
            for category in categories:
                if category.parentCategoryId == parent_id:
                    if depth == 0:
                        node = Text("●", style=category.color)
                    else:
                        node = Text(
                            " " * (depth - 1)
                            + ("└" if is_last(category, parent_id) else "├"),
                            style=category.color,
                        )
                    result.append((category, node))
                    result.extend(build_category_tree(category.id, depth + 1))
            return result

        def is_last(category, parent_id):
            siblings = [cat for cat in categories if cat.parentCategoryId == parent_id]
            return category == siblings[-1]

        return build_category_tree()
    finally:
        session.close()


def get_all_categories_by_freq():
    """Retrieve all categories ordered by the frequency of their usage in records."""
    session = Session()
    try:
        stmt = (
            select(Category, func.count(Category.records).label("record_count"))
            .outerjoin(Category.records)
            .group_by(Category.id)
            .order_by(desc("record_count"))
            .options(joinedload(Category.parentCategory))
            .filter(Category.deletedAt.is_(None))
        )
        return session.execute(stmt).all()
    finally:
        session.close()


def get_category_by_id(category_id):
    """Retrieve a category by its ID."""
    session = Session()
    try:
        stmt = _get_base_categories_query().filter_by(id=category_id)
        return session.scalars(stmt).first()
    finally:
        session.close()


def get_all_categories_records(
    offset: int = 0,
    offset_type: str = "month",
    is_income: bool = True,
    subcategories: bool = False,
    account_id: int = None,
):
    """
    Retrieve all categories with their net income or expenses, sorted by total amount.
    """
    session = Session()
    try:
        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)

        stmt = select(Record).options(joinedload(Record.category))
        if account_id is not None:
            stmt = stmt.filter(Record.accountId == account_id)
        stmt = stmt.filter(
            Record.date >= start_of_period,
            Record.date < end_of_period,
            Record.isIncome == is_income,
        )

        category_totals = {}
        records = session.scalars(stmt).all()
        for record in records:
            split_total = sum(split.amount for split in record.splits)
            record_amount = record.amount - split_total

            if record.category is None:
                continue

            category_id = record.categoryId
            if not subcategories and record.category.parentCategoryId:
                category_id = record.category.parentCategoryId

            if category_id not in category_totals:
                category_totals[category_id] = 0
            category_totals[category_id] += record_amount

        stmt = (
            select(Category)
            .filter(
                Category.id.in_(category_totals.keys()), Category.deletedAt.is_(None)
            )
            .options(joinedload(Category.parentCategory))
        )
        categories = session.scalars(stmt).all()
        for category in categories:
            category.amount = category_totals[category.id]

        categories = [cat for cat in categories if cat.amount != 0]
        categories.sort(key=lambda cat: cat.amount, reverse=True)

        return categories
    finally:
        session.close()


# region Create
def create_category(data):
    """Create a new category."""
    session = Session()
    try:
        new_category = Category(**data)
        session.add(new_category)
        session.commit()
        session.refresh(new_category)
        session.expunge(new_category)
        return new_category
    finally:
        session.close()


# region Update
def update_category(category_id, data):
    """Update a category by its ID."""
    session = Session()
    try:
        category = session.get(Category, category_id)
        if category:
            for key, value in data.items():
                setattr(category, key, value)
            session.commit()
            session.refresh(category)
            session.expunge(category)
        return category
    finally:
        session.close()


# region Delete
def delete_category(category_id):
    """Delete a category by marking it as deleted."""
    session = Session()
    try:
        category = session.get(Category, category_id)
        if category:
            category.deletedAt = datetime.now()
            session.commit()
            session.refresh(category)
            session.expunge(category)
            return True
        return False
    finally:
        session.close()



# region Accounts
from textual import events
from textual.app import ComposeResult
from textual.containers import Container, ScrollableContainer
from textual.widgets import Label, Static

from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal
from bagels.config import CONFIG
from bagels.managers.accounts import (
    create_account,
    delete_account,
    get_all_accounts_with_balance,
    update_account,
)
from bagels.forms.account_forms import AccountForm
from bagels.components.indicators import EmptyIndicator


class AccountMode(ScrollableContainer):
    BINDINGS = [
        (CONFIG.hotkeys.new, "new", "New account"),
        (CONFIG.hotkeys.delete, "delete", "Archive account"),
        (CONFIG.hotkeys.edit, "edit", "Edit account"),
    ]

    can_focus = True

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="accounts-container", classes="module-container"
        )
        super().__setattr__("border_title", "Accounts (using)")
        super().__setattr__(
            "border_subtitle",
            f"{CONFIG.hotkeys.home.select_prev_account} {CONFIG.hotkeys.home.select_next_account}",
        )
        self.page_parent = parent
        self.account_form = AccountForm()

    def on_mount(self) -> None:
        self.rebuild()

    # region Builder
    # -------------- Builder ------------- #

    def rebuild(self) -> None:
        for account in get_all_accounts_with_balance():
            # Update balance
            self.query_one(f"#account-{account.id}-balance").update(
                str(account.balance)
            )

            # Update account container classes
            account_container = self.query_one(f"#account-{account.id}-container")
            selected = (
                "selected"
                if self.page_parent.mode["accountId"]["default_value"] == account.id
                else ""
            )
            account_container.classes = f"account-container {selected}"

            # Update name/description label
            name_label = self.query_one(f"#account-{account.id}-name")
            name_label.update(account.name)
            description_label: Label = self.query_one(
                f"#account-{account.id}-description"
            )
            description_label.update(account.description)
            if account.description == "" or account.description is None:
                description_label.add_class("none")
            else:
                description_label.remove_class("none")

            # Scroll to selected account
            if selected:
                self.scroll_to_widget(account_container)

    # region Callbacks
    # ------------- Callbacks ------------ #

    def on_key(self, event: events.Key) -> None:
        if event.key == "up":
            self.page_parent.action_select_prev_account()
        elif event.key == "down":
            self.page_parent.action_select_next_account()

    # region cud
    # ---------------- cud --------------- #

    def action_new(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_account(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                self.app.notify(
                    title="Success",
                    message=f"Account {result['name']} created",
                    severity="information",
                    timeout=3,
                )
                self.app.refresh(layout=True, recompose=True)  # the big button

        account_form = self.account_form.get_form()
        self.app.push_screen(
            InputModal("New Account", account_form), callback=check_result
        )

    def action_edit(self) -> None:
        id = self.page_parent.mode["accountId"]["default_value"]

        def check_result(result) -> None:
            if result:
                try:
                    update_account(id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                self.app.notify(
                    title="Success",
                    message=f"Account {result['name']} updated",
                    severity="information",
                    timeout=3,
                )
                self.page_parent.rebuild()

        if id:
            filled_account_form = self.account_form.get_filled_form(id)
            self.app.push_screen(
                InputModal("Edit Account", filled_account_form), callback=check_result
            )
        else:
            self.app.notify(
                title="Error",
                message="No account selected",
                severity="error",
                timeout=3,
            )

    def action_delete(self) -> None:
        id = self.page_parent.mode["accountId"]["default_value"]

        def check_delete(result) -> None:
            if result:
                delete_account(id)
                self.app.notify(
                    title="Success",
                    message=f"Archived account",
                    severity="information",
                    timeout=3,
                )
                self.app.refresh(layout=True, recompose=True)  # the big button

        if id:
            self.app.push_screen(
                ConfirmationModal("Are you sure you want to archive this account?"),
                check_delete,
            )
        else:
            self.app.notify(
                title="Error",
                message="No account selected",
                severity="error",
                timeout=3,
            )
        pass

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        accounts = get_all_accounts_with_balance()
        if not accounts:
            yield EmptyIndicator("No accounts")
            return

        for account in accounts:
            with Container(
                id=f"account-{account.id}-container", classes="account-container"
            ):
                with Container(classes="left-container"):
                    yield Label(
                        "",  # Will be populated in rebuild()
                        classes="name",
                        id=f"account-{account.id}-name",
                    )
                    yield Label(
                        "",  # Will be populated in rebuild()
                        classes="description",
                        id=f"account-{account.id}-description",
                    )
                yield Label("", classes="balance", id=f"account-{account.id}-balance")



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from bagels.models.database.db import Base
from bagels.models.person import Person
from bagels.managers import persons

@pytest.fixture(scope="function")
def test_db():
    # Create in-memory SQLite database
    engine = create_engine("sqlite:///:memory:")
    
    # Create all tables
    Base.metadata.create_all(engine)
    
    # Create a new session factory bound to the engine
    persons.Session = sessionmaker(bind=engine)
    
    yield engine
    
    # Clean up
    Base.metadata.drop_all(engine)

def test_create_person(test_db):
    # Test data
    person_data = {"name": "John Doe"}
    
    # Create person
    new_person = persons.create_person(person_data)
    
    # Assertions
    assert new_person is not None
    assert new_person.name == "John Doe"
    assert new_person.id is not None
    assert new_person.createdAt is not None
    assert new_person.updatedAt is not None

def test_get_all_persons(test_db):
    # Create test data
    person_data1 = {"name": "John Doe"}
    person_data2 = {"name": "Jane Smith"}
    
    persons.create_person(person_data1)
    persons.create_person(person_data2)
    
    # Get all persons
    all_persons = persons.get_all_persons()
    
    # Assertions
    assert len(all_persons) == 2
    assert any(p.name == "John Doe" for p in all_persons)
    assert any(p.name == "Jane Smith" for p in all_persons)

def test_get_person_by_id(test_db):
    # Create test data
    person_data = {"name": "John Doe"}
    new_person = persons.create_person(person_data)
    
    # Get person by ID
    retrieved_person = persons.get_person_by_id(new_person.id)
    
    # Assertions
    assert retrieved_person is not None
    assert retrieved_person.id == new_person.id
    assert retrieved_person.name == "John Doe"

def test_get_nonexistent_person(test_db):
    # Try to get a person with non-existent ID
    retrieved_person = persons.get_person_by_id(999)
    
    # Assertions
    assert retrieved_person is None

def test_update_person(test_db):
    # Create test data
    person_data = {"name": "John Doe"}
    new_person = persons.create_person(person_data)
    
    # Update data
    update_data = {"name": "John Smith"}
    updated_person = persons.update_person(new_person.id, update_data)
    
    # Assertions
    assert updated_person is not None
    assert updated_person.name == "John Smith"
    assert updated_person.id == new_person.id

def test_update_nonexistent_person(test_db):
    # Try to update non-existent person
    update_data = {"name": "John Smith"}
    updated_person = persons.update_person(999, update_data)
    
    # Assertions
    assert updated_person is None

def test_delete_person(test_db):
    # Create test data
    person_data = {"name": "John Doe"}
    new_person = persons.create_person(person_data)
    
    # Delete person
    result = persons.delete_person(new_person.id)
    
    # Assertions
    assert result is True
    assert persons.get_person_by_id(new_person.id) is None

def test_delete_nonexistent_person(test_db):
    # Try to delete non-existent person
    result = persons.delete_person(999)
    
    # Assertions
    assert result is False



import pytest
from datetime import datetime, timedelta
from freezegun import freeze_time
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from bagels.models.database.db import Base
from bagels.models.account import Account
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.models.person import Person
from bagels.models.category import Category, Nature
from bagels.managers import utils
from bagels.config import CONFIG

# Test fixtures
@pytest.fixture(scope="function")
def engine():
    """Create a test-specific database engine."""
    return create_engine("sqlite:///:memory:")

@pytest.fixture(scope="function")
def setup_test_database(engine):
    """Create all tables in the test database."""
    Base.metadata.create_all(engine)
    yield
    Base.metadata.drop_all(engine)

@pytest.fixture(scope="function")
def session(engine, setup_test_database):
    """Create a new session for a test."""
    Session = sessionmaker(bind=engine)
    session = Session()
    yield session
    session.close()

@pytest.fixture
def test_data(session):
    """Create test accounts, categories, and people."""
    # Create test accounts
    account1 = Account(name="Account 1", beginningBalance=1000.0)
    account2 = Account(name="Account 2", beginningBalance=500.0)
    session.add_all([account1, account2])
    
    # Create a test category
    category = Category(name="Test Category", nature=Nature.NEED, color="#FF0000")
    session.add(category)
    
    # Create a test person
    person = Person(name="Test Person")
    session.add(person)
    
    session.commit()
    
    return {
        "account1": account1,
        "account2": account2,
        "category": category,
        "person": person
    }

# Test period calculations
@freeze_time("2024-02-15")  # Freeze time to February 15, 2024
def test_get_start_end_of_year():
    """Test year period calculations."""
    # Current year
    start, end = utils._get_start_end_of_year(0)
    assert start == datetime(2024, 1, 1, 0, 0, 0)
    assert end == datetime(2024, 12, 31, 23, 59, 59)
    
    # Previous year
    start, end = utils._get_start_end_of_year(-1)
    assert start == datetime(2023, 1, 1, 0, 0, 0)
    assert end == datetime(2023, 12, 31, 23, 59, 59)
    
    # Next year
    start, end = utils._get_start_end_of_year(1)
    assert start == datetime(2025, 1, 1, 0, 0, 0)
    assert end == datetime(2025, 12, 31, 23, 59, 59)

@freeze_time("2024-02-15")  # Freeze time to February 15, 2024
def test_get_start_end_of_month():
    """Test month period calculations."""
    # Current month
    start, end = utils._get_start_end_of_month(0)
    assert start == datetime(2024, 2, 1, 0, 0, 0)
    assert end == datetime(2024, 2, 29, 23, 59, 59)  # 2024 is a leap year
    
    # Previous month
    start, end = utils._get_start_end_of_month(-1)
    assert start == datetime(2024, 1, 1, 0, 0, 0)
    assert end == datetime(2024, 1, 31, 23, 59, 59)
    
    # Next month
    start, end = utils._get_start_end_of_month(1)
    assert start == datetime(2024, 3, 1, 0, 0, 0)
    assert end == datetime(2024, 3, 31, 23, 59, 59)

@freeze_time("2024-02-15")  # Freeze time to February 15, 2024
def test_get_start_end_of_week():
    """Test week period calculations."""
    # Set first day of week to Monday (0)
    CONFIG.defaults.first_day_of_week = 0
    
    # Current week
    start, end = utils._get_start_end_of_week(0)
    assert start == datetime(2024, 2, 12, 0, 0, 0)  # Monday
    assert end == datetime(2024, 2, 18, 23, 59, 59)  # Sunday
    
    # Previous week
    start, end = utils._get_start_end_of_week(-1)
    assert start == datetime(2024, 2, 5, 0, 0, 0)
    assert end == datetime(2024, 2, 11, 23, 59, 59)
    
    # Next week
    start, end = utils._get_start_end_of_week(1)
    assert start == datetime(2024, 2, 19, 0, 0, 0)
    assert end == datetime(2024, 2, 25, 23, 59, 59)

@freeze_time("2024-02-15")  # Freeze time to February 15, 2024
def test_get_start_end_of_day():
    """Test day period calculations."""
    # Current day
    start, end = utils._get_start_end_of_day(0)
    assert start == datetime(2024, 2, 15, 0, 0, 0)
    assert end == datetime(2024, 2, 15, 23, 59, 59)
    
    # Previous day
    start, end = utils._get_start_end_of_day(-1)
    assert start == datetime(2024, 2, 14, 0, 0, 0)
    assert end == datetime(2024, 2, 14, 23, 59, 59)
    
    # Next day
    start, end = utils._get_start_end_of_day(1)
    assert start == datetime(2024, 2, 16, 0, 0, 0)
    assert end == datetime(2024, 2, 16, 23, 59, 59)

@freeze_time("2024-02-15")  # Freeze time to February 15, 2024
def test_get_start_end_of_period():
    """Test generic period calculations."""
    # Test year
    start, end = utils.get_start_end_of_period(0, "year")
    assert start == datetime(2024, 1, 1, 0, 0, 0)
    assert end == datetime(2024, 12, 31, 23, 59, 59)
    
    # Test month
    start, end = utils.get_start_end_of_period(0, "month")
    assert start == datetime(2024, 2, 1, 0, 0, 0)
    assert end == datetime(2024, 2, 29, 23, 59, 59)
    
    # Test week
    CONFIG.defaults.first_day_of_week = 0  # Monday
    start, end = utils.get_start_end_of_period(0, "week")
    assert start == datetime(2024, 2, 12, 0, 0, 0)
    assert end == datetime(2024, 2, 18, 23, 59, 59)
    
    # Test day
    start, end = utils.get_start_end_of_period(0, "day")
    assert start == datetime(2024, 2, 15, 0, 0, 0)
    assert end == datetime(2024, 2, 15, 23, 59, 59)

# Test figure calculations
@freeze_time("2024-02-15")
def test_get_period_figures(session, test_data):
    """Test period figure calculations."""
    # Create test records
    income_record = Record(
        label="Test Income",
        amount=200.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=True,
        date=datetime(2024, 2, 15)
    )
    expense_record = Record(
        label="Test Expense",
        amount=150.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime(2024, 2, 15)
    )
    transfer_record = Record(
        label="Test Transfer",
        amount=300.0,
        accountId=test_data["account1"].id,
        isTransfer=True,
        transferToAccountId=test_data["account2"].id,
        date=datetime(2024, 2, 15)
    )
    split_record = Record(
        label="Test Split",
        amount=400.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime(2024, 2, 15)
    )
    session.add_all([income_record, expense_record, transfer_record, split_record])
    session.flush()
    
    # Create split
    split = Split(
        recordId=split_record.id,
        amount=200.0,
        personId=test_data["person"].id,
        isPaid=True,
        accountId=test_data["account1"].id,
        paidDate=datetime(2024, 2, 15)
    )
    session.add(split)
    session.commit()
    
    # Test income figures
    income = utils.get_period_figures(
        accountId=test_data["account1"].id,
        offset_type="month",
        offset=0,
        isIncome=True,
        session=session
    )
    assert income == 200.0  # Only the income record
    
    # Test expense figures
    expenses = utils.get_period_figures(
        accountId=test_data["account1"].id,
        offset_type="month",
        offset=0,
        isIncome=False,
        session=session
    )
    assert expenses == 350.0  # 150 (expense) + 200 (split expense after paid split)

# Test average calculations
def test_get_days_in_period():
    """Test days in period calculations."""
    with freeze_time("2024-02-15"):
        # Test month (February 2024 - leap year)
        days = utils._get_days_in_period(0, "month")
        assert days == 29
        
        # Test year
        days = utils._get_days_in_period(0, "year")
        assert days == 366  # 2024 is a leap year
        
        # Test week
        days = utils._get_days_in_period(0, "week")
        assert days == 7
        
        # Test day
        days = utils._get_days_in_period(0, "day")
        assert days == 1

def test_get_period_average():
    """Test period average calculations."""
    with freeze_time("2024-02-15"):
        # Test month average (February 2024)
        avg = utils.get_period_average(net=290, offset=0, offset_type="month")
        assert avg == 10.0  # 290 / 29 days
        
        # Test week average
        avg = utils.get_period_average(net=70, offset=0, offset_type="week")
        assert avg == 10.0  # 70 / 7 days
        
        # Test day average
        avg = utils.get_period_average(net=10, offset=0, offset_type="day")
        assert avg == 10.0  # 10 / 1 day



from pydantic import BaseModel, Field
from textual.design import ColorSystem


class Theme(BaseModel):
    primary: str
    secondary: str | None = None
    warning: str | None = None
    error: str | None = None
    success: str | None = None
    accent: str | None = None
    foreground: str | None = None
    background: str | None = None
    surface: str | None = None
    panel: str | None = None
    boost: str | None = None
    dark: bool = True
    luminosity_spread: float = 0.15
    text_alpha: float = 0.95
    variables: dict[str, str] = Field(default_factory=dict)

    def to_color_system(self) -> ColorSystem:
        """Convert this theme to a ColorSystem."""
        return ColorSystem(**self.model_dump())


BUILTIN_THEMES: dict[str, Theme] = {
    "dark": Theme(
        name="textual-dark",
        primary="#0178D4",
        secondary="#004578",
        accent="#ffa62b",
        warning="#ffa62b",
        error="#ba3c5b",
        success="#4EBF71",
        foreground="#e0e0e0",
    ),
    "galaxy": Theme(
        name="galaxy",
        primary="#8A2BE2",  # Improved Deep Magenta (Blueviolet)
        secondary="#a684e8",
        warning="#FFD700",  # Gold, more visible than orange
        error="#FF4500",  # OrangeRed, vibrant but less harsh than pure red
        success="#00FA9A",  # Medium Spring Green, kept for vibrancy
        accent="#FF69B4",  # Hot Pink, for a pop of color
        dark=True,
        background="#0F0F1F",  # Very Dark Blue, almost black
        surface="#1E1E3F",  # Dark Blue-Purple
        panel="#2D2B55",  # Slightly Lighter Blue-Purple
    ),
    "alpine": Theme(
        name="alpine",
        primary="#4A90E2",  # Clear Sky Blue
        secondary="#81A1C1",  # Misty Blue
        warning="#EBCB8B",  # Soft Sunlight
        error="#BF616A",  # Muted Red
        success="#A3BE8C",  # Alpine Meadow Green
        accent="#5E81AC",  # Mountain Lake Blue
        dark=True,
        background="#262b35",  # Dark Slate Grey
        surface="#3B4252",  # Darker Blue-Grey
        panel="#434C5E",  # Lighter Blue-Grey
    ),
    "cobalt": Theme(
        name="cobalt",
        primary="#334D5C",  # Deep Cobalt Blue
        secondary="#4878A6",  # Slate Blue
        warning="#FFAA22",  # Amber, suitable for warnings related to primary
        error="#E63946",  # Red, universally recognized for errors
        success="#4CAF50",  # Green, commonly used for success indication
        accent="#D94E64",  # Candy Apple Red
        dark=True,
        surface="#27343B",  # Dark Lead
        panel="#2D3E46",  # Storm Gray
        background="#1F262A",  # Charcoal
    ),
    "hacker": Theme(
        name="hacker",
        primary="#00FF00",  # Bright Green (Lime)
        secondary="#32CD32",  # Lime Green
        warning="#ADFF2F",  # Green Yellow
        error="#FF4500",  # Orange Red (for contrast)
        success="#00FA9A",  # Medium Spring Green
        accent="#39FF14",  # Neon Green
        dark=True,
        background="#0D0D0D",  # Almost Black
        surface="#1A1A1A",  # Very Dark Gray
        panel="#2A2A2A",  # Dark Gray
    ),
    "nord": Theme(
        name="nord",
        primary="#88C0D0",
        secondary="#81A1C1",
        accent="#B48EAD",
        foreground="#D8DEE9",
        background="#2E3440",
        success="#A3BE8C",
        warning="#EBCB8B",
        error="#BF616A",
        surface="#3B4252",
        panel="#434C5E",
        variables={
            "block-cursor-background": "#88C0D0",
            "block-cursor-foreground": "#2E3440",
            "block-cursor-text-style": "none",
            "footer-key-foreground": "#88C0D0",
            "input-selection-background": "#81a1c1 35%",
            "button-color-foreground": "#2E3440",
            "button-focus-text-style": "reverse",
        },
    ),
    "gruvbox": Theme(
        name="gruvbox",
        primary="#85A598",
        secondary="#A89A85",
        warning="#fe8019",
        error="#fb4934",
        success="#b8bb26",
        accent="#fabd2f",
        foreground="#fbf1c7",
        background="#282828",
        surface="#3c3836",
        panel="#504945",
        variables={
            "block-cursor-foreground": "#fbf1c7",
            "input-selection-background": "#689d6a40",
            "button-color-foreground": "#282828",
        },
    ),
    "catppuccin-mocha": Theme(
        name="catppuccin-mocha",
        primary="#F5C2E7",
        secondary="#cba6f7",
        warning="#FAE3B0",
        error="#F28FAD",
        success="#ABE9B3",
        accent="#fab387",
        foreground="#cdd6f4",
        background="#181825",
        surface="#313244",
        panel="#45475a",
        variables={
            "input-cursor-foreground": "#11111b",
            "input-cursor-background": "#f5e0dc",
            "input-selection-background": "#9399b2 30%",
            "border": "#b4befe",
            "border-blurred": "#585b70",
            "footer-background": "#45475a",
            "block-cursor-foreground": "#1e1e2e",
            "block-cursor-text-style": "none",
            "button-color-foreground": "#181825",
        },
    ),
    "dracula": Theme(
        name="dracula",
        primary="#BD93F9",
        secondary="#6272A4",
        warning="#FFB86C",
        error="#FF5555",
        success="#50FA7B",
        accent="#FF79C6",
        background="#282A36",
        surface="#2B2E3B",
        panel="#313442",
        foreground="#F8F8F2",
        variables={
            "button-color-foreground": "#282A36",
        },
    ),
    "tokyo-night": Theme(
        name="tokyo-night",
        primary="#BB9AF7",
        secondary="#7AA2F7",
        warning="#E0AF68",  # Yellow
        error="#F7768E",  # Red
        success="#9ECE6A",  # Green
        accent="#FF9E64",  # Orange
        foreground="#a9b1d6",
        background="#1A1B26",  # Background
        surface="#24283B",  # Surface
        panel="#414868",  # Panel
        variables={
            "button-color-foreground": "#24283B",
        },
    ),
    "flexoki": Theme(
        name="flexoki",
        primary="#205EA6",  # blue
        secondary="#24837B",  # cyan
        warning="#AD8301",  # yellow
        error="#AF3029",  # red
        success="#66800B",  # green
        accent="#9B76C8",  # purple light
        background="#100F0F",  # base.black
        surface="#1C1B1A",  # base.950
        panel="#282726",  # base.900
        foreground="#FFFCF0",  # base.paper
        variables={
            "input-cursor-foreground": "#5E409D",
            "input-cursor-background": "#FFFCF0",
            "input-selection-background": "#6F6E69 35%",  # base.600 with opacity
            "button-color-foreground": "#FFFCF0",
        },
    ),
    # "textual-light": Theme(
    #     name="textual-light",
    #     primary="#004578",
    #     secondary="#0178D4",
    #     accent="#ffa62b",
    #     warning="#ffa62b",
    #     error="#ba3c5b",
    #     success="#4EBF71",
    #     surface="#D8D8D8",
    #     panel="#D0D0D0",
    #     background="#E0E0E0",
    #     dark=False,
    #     variables={
    #         "footer-key-foreground": "#0178D4",
    #     },
    # ),
    # "catppuccin-latte": Theme(
    #     name="catppuccin-latte",
    #     secondary="#DC8A78",
    #     primary="#8839EF",
    #     warning="#DF8E1D",
    #     error="#D20F39",
    #     success="#40A02B",
    #     accent="#FE640B",
    #     foreground="#4C4F69",
    #     background="#EFF1F5",
    #     surface="#E6E9EF",
    #     panel="#CCD0DA",
    #     dark=False,
    #     variables={
    #         "button-color-foreground": "#EFF1F5",
    #     },
    # ),
    # "solarized-light": Theme(
    #     name="solarized-light",
    #     primary="#268bd2",
    #     secondary="#2aa198",
    #     warning="#cb4b16",
    #     error="#dc322f",
    #     success="#859900",
    #     accent="#6c71c4",
    #     foreground="#586e75",
    #     background="#fdf6e3",
    #     surface="#eee8d5",
    #     panel="#eee8d5",
    #     dark=False,
    #     variables={
    #         "button-color-foreground": "#fdf6e3",
    #         "footer-background": "#268bd2",
    #         "footer-key-foreground": "#fdf6e3",
    #         "footer-description-foreground": "#fdf6e3",
    #     },
    # ),
}



import copy
from rich.text import Text

from bagels.managers.accounts import get_all_accounts_with_balance
from bagels.managers.categories import get_all_categories_by_freq
from bagels.managers.record_templates import get_template_by_id
from bagels.forms.form import Form, FormField, Option, Options


class RecordTemplateForm:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # ------------ Blueprints ------------ #

    FORM = Form(
        fields=[
            FormField(
                placeholder="Template label",
                title="Label",
                key="label",
                type="string",
                is_required=True,
            ),
            FormField(
                title="Category",
                key="categoryId",
                type="autocomplete",
                options=Options(),
                is_required=True,
                placeholder="Select Category",
            ),
            FormField(
                placeholder="0.00",
                title="Amount",
                key="amount",
                type="number",
                min=0,
                is_required=True,
            ),
            FormField(
                title="Account",
                key="accountId",
                type="autocomplete",
                options=Options(),
                is_required=True,
                placeholder="Select Account",
            ),
            FormField(
                title="Type",
                key="isIncome",
                type="boolean",
                labels=["Expense", "Income"],
                default_value=False,
            ),
        ]
    )

    # ----------------- - ---------------- #

    def __init__(self):
        self._populate_form_options()

    # -------------- Helpers ------------- #

    def _populate_form_options(self):
        accounts = get_all_accounts_with_balance()
        self.FORM.fields[3].options = Options(
            items=[
                Option(
                    text=account.name,
                    value=account.id,
                    postfix=Text(f"{account.balance}", style="yellow"),
                )
                for account in accounts
            ]
        )

        categories = get_all_categories_by_freq()
        self.FORM.fields[1].options = Options(
            items=[
                Option(
                    text=category.name,
                    value=category.id,
                    prefix=Text("●", style=category.color),
                    postfix=(
                        Text(
                            (
                                f"↪ {category.parentCategory.name}"
                                if category.parentCategory
                                else ""
                            ),
                            style=category.parentCategory.color,
                        )
                        if category.parentCategory
                        else ""
                    ),
                )
                for category, _ in categories
            ]
        )

    # ------------- Builders ------------- #

    def get_filled_form(self, templateId: int) -> list:
        """Return a copy of the form with values from the record"""
        filled_form = copy.deepcopy(self.FORM)
        template = get_template_by_id(templateId)

        for field in filled_form.fields:
            fieldKey = field.key
            value = getattr(template, fieldKey)

            match fieldKey:
                case "isIncome":
                    field.default_value = value
                case "categoryId":
                    field.default_value = template.category.id
                    field.default_value_text = template.category.name
                case "accountId":
                    field.default_value = template.account.id
                    field.default_value_text = template.account.name
                case _:
                    field.default_value = str(value) if value is not None else ""

        return filled_form

    def get_form(self):
        """Return the base form with default values"""
        form = copy.deepcopy(self.FORM)
        return form



import pytest
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from bagels.models.database.db import Base
from bagels.models.category import Category, Nature
from bagels.managers import categories

@pytest.fixture(scope="function")
def test_db():
    # Create in-memory SQLite database
    engine = create_engine("sqlite:///:memory:")
    
    # Create all tables
    Base.metadata.create_all(engine)
    
    # Create a new session factory bound to the engine
    categories.Session = sessionmaker(bind=engine)
    
    yield engine
    
    # Clean up
    Base.metadata.drop_all(engine)

def test_create_category(test_db):
    # Test data
    category_data = {
        "name": "Test Category",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    
    # Create category
    new_category = categories.create_category(category_data)
    
    # Assertions
    assert new_category is not None
    assert new_category.name == "Test Category"
    assert new_category.nature == Nature.WANT
    assert new_category.color == "#FF0000"
    assert new_category.id is not None
    assert new_category.createdAt is not None
    assert new_category.updatedAt is not None
    assert new_category.deletedAt is None
    assert new_category.parentCategoryId is None

def test_create_subcategory(test_db):
    # Create parent category first
    parent_data = {
        "name": "Parent Category",
        "nature": Nature.NEED,
        "color": "#00FF00"
    }
    parent = categories.create_category(parent_data)
    
    # Create subcategory
    child_data = {
        "name": "Child Category",
        "nature": Nature.WANT,
        "color": "#0000FF",
        "parentCategoryId": parent.id
    }
    child = categories.create_category(child_data)
    
    # Assertions
    assert child is not None
    assert child.parentCategoryId == parent.id
    assert child.name == "Child Category"

def test_get_categories_count(test_db):
    # Create test categories
    category_data1 = {
        "name": "Category 1",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    category_data2 = {
        "name": "Category 2",
        "nature": Nature.NEED,
        "color": "#00FF00"
    }
    
    categories.create_category(category_data1)
    categories.create_category(category_data2)
    
    # Test count
    assert categories.get_categories_count() == 2

def test_get_all_categories_tree(test_db):
    # Create test categories with hierarchy
    parent_data = {
        "name": "Parent",
        "nature": Nature.NEED,
        "color": "#FF0000"
    }
    parent = categories.create_category(parent_data)
    
    child_data = {
        "name": "Child",
        "nature": Nature.WANT,
        "color": "#00FF00",
        "parentCategoryId": parent.id
    }
    categories.create_category(child_data)
    
    # Get tree
    tree = categories.get_all_categories_tree()
    
    # Assertions
    assert len(tree) == 2  # Parent and child nodes
    assert tree[0][0].name == "Parent"  # First node is parent
    assert tree[1][0].name == "Child"  # Second node is child
    assert tree[1][0].parentCategoryId == parent.id

def test_get_all_categories_by_freq(test_db):
    # Create test categories
    category_data = {
        "name": "Test Category",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    categories.create_category(category_data)
    
    # Get categories by frequency
    result = categories.get_all_categories_by_freq()
    
    # Assertions
    assert len(result) == 1
    category, count = result[0]
    assert category.name == "Test Category"
    assert count == 0  # No records yet

def test_get_category_by_id(test_db):
    # Create test category
    category_data = {
        "name": "Test Category",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    new_category = categories.create_category(category_data)
    
    # Get category by ID
    retrieved_category = categories.get_category_by_id(new_category.id)
    
    # Assertions
    assert retrieved_category is not None
    assert retrieved_category.id == new_category.id
    assert retrieved_category.name == "Test Category"

def test_get_nonexistent_category(test_db):
    # Try to get a category with non-existent ID
    retrieved_category = categories.get_category_by_id(999)
    
    # Assertions
    assert retrieved_category is None

def test_update_category(test_db):
    # Create test category
    category_data = {
        "name": "Test Category",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    new_category = categories.create_category(category_data)
    
    # Update data
    update_data = {
        "name": "Updated Category",
        "nature": Nature.NEED,
        "color": "#00FF00"
    }
    updated_category = categories.update_category(new_category.id, update_data)
    
    # Assertions
    assert updated_category is not None
    assert updated_category.name == "Updated Category"
    assert updated_category.nature == Nature.NEED
    assert updated_category.color == "#00FF00"
    assert updated_category.id == new_category.id

def test_update_nonexistent_category(test_db):
    # Try to update non-existent category
    update_data = {"name": "Updated Category"}
    updated_category = categories.update_category(999, update_data)
    
    # Assertions
    assert updated_category is None

def test_delete_category(test_db):
    # Create test category
    category_data = {
        "name": "Test Category",
        "nature": Nature.WANT,
        "color": "#FF0000"
    }
    new_category = categories.create_category(category_data)
    
    # Delete category
    result = categories.delete_category(new_category.id)
    
    # Assertions
    assert result is True
    
    # Verify the category is marked as deleted
    deleted_category = categories.get_category_by_id(new_category.id)
    assert deleted_category is None  # Should not be returned by get_category_by_id
    
    # Verify it's not returned in counts
    assert categories.get_categories_count() == 0

def test_delete_nonexistent_category(test_db):
    # Try to delete non-existent category
    result = categories.delete_category(999)
    
    # Assertions
    assert result is False




from textual import events
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import (
    Label,
)



class ConfirmationModal(ModalScreen):
    def __init__(self, message: str, *args, **kwargs):
        super().__init__(id="confirmation-modal-screen", *args, **kwargs)
        self.message = message

    def compose(self) -> ComposeResult:
        dialog = Container(classes="dialog")
        dialog.border_title = "Alert"
        with dialog:
            yield Label(self.message)

    def on_key(self, event: events.Key):
        if event.key == "enter":
            self.dismiss(True)
        elif event.key == "escape":
            self.dismiss(False)



from pathlib import Path
import yaml

from bagels.models.account import Account
from bagels.models.database.app import db_engine
from bagels.models.person import Person
from bagels.models.record import Record
from bagels.models.record_template import RecordTemplate
from bagels.models.split import Split
from sqlalchemy.orm import sessionmaker
from sqlalchemy import select, func

Session = sessionmaker(bind=db_engine)


def create_sample_entries():
    yaml_path = Path(__file__).parent.parent / "static" / "sample_entries.yaml"

    with open(yaml_path, "r") as file:
        sample_entries = yaml.safe_load(file)

    session = Session()
    try:
        # Create accounts
        accounts = {}
        for account_data in sample_entries["accounts"]:
            account = Account(**account_data)
            session.add(account)
            session.flush()
            accounts[account.id] = account

        # Create people
        people = {}
        for person_data in sample_entries["people"]:
            person = Person(**person_data)
            session.add(person)
            session.flush()
            people[person.id] = person

        # Create records
        for record_data in sample_entries["records"]:
            splits_data = record_data.pop("splits", None)
            record = Record(**record_data)
            session.add(record)
            session.flush()

            if splits_data:
                for split_data in splits_data:
                    split = Split(recordId=record.id, **split_data)
                    session.add(split)

        # Create record templates
        for template_data in sample_entries["record_templates"]:
            # Find the current maximum order
            max_order = (
                session.execute(
                    select(func.max(RecordTemplate.order))
                ).scalar_one_or_none()
                or 0
            )

            # Increment and set the order
            max_order += 1
            template_data["order"] = max_order
            template = RecordTemplate(**template_data)
            session.add(template)

        session.commit()
    finally:
        session.close()



from datetime import datetime, timedelta

from bagels.config import CONFIG
from bagels.models.database.app import db_engine
from bagels.models.record import Record
from sqlalchemy.orm import sessionmaker

Session = sessionmaker(bind=db_engine)


# region period
# -------------- period -------------- #
def _get_start_end_of_year(offset: int = 0):
    now = datetime.now()
    target_year = now.year + offset
    start_of_year = datetime(target_year, 1, 1, 0, 0, 0)
    end_of_year = datetime(target_year, 12, 31, 23, 59, 59)
    return start_of_year, end_of_year


def _get_start_end_of_month(offset: int = 0):
    now = datetime.now()
    # Calculate target month and year
    target_month = now.month + offset
    target_year = now.year + (target_month - 1) // 12
    target_month = ((target_month - 1) % 12) + 1

    # Calculate next month and year for end date
    next_month = target_month + 1
    next_year = target_year + (next_month - 1) // 12
    next_month = ((next_month - 1) % 12) + 1

    start_of_month = datetime(target_year, target_month, 1, 0, 0, 0)
    end_of_month = datetime(next_year, next_month, 1, 0, 0, 0) - timedelta(seconds=1)

    return start_of_month, end_of_month


def _get_start_end_of_week(offset: int = 0):
    now = datetime.now()
    # Apply offset in weeks
    now = now + timedelta(weeks=offset)
    first_day_of_week = CONFIG.defaults.first_day_of_week
    start_of_week = (
        now - timedelta(days=(now.weekday() - first_day_of_week) % 7)
    ).replace(hour=0, minute=0, second=0, microsecond=0)
    end_of_week = (start_of_week + timedelta(days=6)).replace(
        hour=23, minute=59, second=59
    )
    return start_of_week, end_of_week


def _get_start_end_of_day(offset: int = 0):
    now = datetime.now()
    # Apply offset in days
    now = now + timedelta(days=offset)
    start_of_day = now.replace(hour=0, minute=0, second=0, microsecond=0)
    end_of_day = now.replace(hour=23, minute=59, second=59)
    return start_of_day, end_of_day


def get_start_end_of_period(offset: int = 0, offset_type: str = "month"):
    match offset_type:
        case "year":
            return _get_start_end_of_year(offset)
        case "month":
            return _get_start_end_of_month(offset)
        case "week":
            return _get_start_end_of_week(offset)
        case "day":
            return _get_start_end_of_day(offset)


# region figure
# -------------- figure -------------- #


def get_period_figures(accountId=None, offset_type=None, offset=None, isIncome=None, session=None):
    """Returns the income / expense for a given period.

    Rules:
    - Filter applies to records only by the "date" column, and splits should always be considered by their associated record.
    - Income and expenses should be calculated from record less their splits, regardless of the account of the split.
    - Transfers are not income or expenses, but should be included in the net total.

    Args:
        accountId (int): The ID of the account to filter by. (Optional)
        offset_type (str): The type of period to filter by.
        offset (int): The offset from the current period.
        isIncome (bool): Whether to filter by income or expense.
        session (Session, optional): SQLAlchemy session to use. If None, creates a new session.
    """
    if session is None:
        session = Session()
        should_close = True
    else:
        should_close = False

    try:
        query = session.query(Record)

        # Filter by account if specified
        if accountId is not None:
            query = query.filter(Record.accountId == accountId)

        # Filter by date period if specified
        if offset_type is not None and offset is not None:
            start_of_period, end_of_period = get_start_end_of_period(
                offset, offset_type
            )
            query = query.filter(
                Record.date >= start_of_period, Record.date < end_of_period
            )

        # Calculate net amount
        total = 0
        records = query.all()

        for record in records:
            # Skip records that are marked as transfers if filtering by income type
            if isIncome is not None and record.isTransfer:
                continue

            # Skip records that don't match income filter if specified
            if isIncome is not None and record.isIncome != isIncome:
                continue

            # Calculate amount after subtracting splits
            split_total = sum(split.amount for split in record.splits)
            record_amount = record.amount - split_total

            # For transfers, only consider the direction
            if record.isTransfer:
                if record.accountId == accountId:
                    total -= record_amount  # Money leaving this account
                else:
                    total += record_amount  # Money entering this account
            else:
                # For regular income/expenses
                if record.isIncome:
                    total += record_amount
                else:
                    total -= record_amount

        return abs(round(total, 2))
    finally:
        if should_close:
            session.close()


# region average
# -------------- average ------------- #


def _get_days_in_period(offset: int = 0, offset_type: str = "month"):
    start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)
    days = (end_of_period - start_of_period).days + 1
    return days


def get_period_average(net: int = 0, offset: int = 0, offset_type: str = "month"):
    days = _get_days_in_period(offset, offset_type)
    return round(net / days, 2)



from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal, HorizontalScroll
from textual.widgets import Label, Static

from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal
from bagels.config import CONFIG
from bagels.models.record_template import RecordTemplate
from bagels.managers.record_templates import (
    create_template,
    delete_template,
    get_adjacent_template,
    get_all_templates,
    swap_template_order,
    update_template,
)
from bagels.managers.records import create_record
from bagels.forms.recordtemplate_forms import RecordTemplateForm


class Templates(Static):
    can_focus = True

    BINDINGS = [
        Binding(CONFIG.hotkeys.new, "new_template", "New Template"),
        Binding(CONFIG.hotkeys.edit, "edit_template", "Edit Template"),
        Binding(CONFIG.hotkeys.delete, "delete_template", "Delete Template"),
        Binding("left", "swap_previous", "Swap left"),
        Binding("right", "swap_next", "Swap right"),
    ]

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="templates-container", classes="module-container"
        )
        super().__setattr__("border_title", "Templates")
        super().__setattr__("border_subtitle", "1 - 9")
        self.page_parent = parent
        self.templates: list[RecordTemplate] = []
        self.template_form = RecordTemplateForm()
        self.selected_template_id = None

    def on_mount(self) -> None:
        self.rebuild()

    # region Builder
    # -------------- Builder ------------- #

    def _create_templates_widgets(self, container: Container) -> None:
        if len(self.templates) == 0:
            widget = Label("No templates. Jump here to create one.", classes="empty")
            container.compose_add_child(widget)
            return container
        for index, template in enumerate(self.templates):
            if index > 8:
                break
            color = template.category.color
            widget = Container(
                Label(
                    f"[{color}]{CONFIG.symbols.category_color}[/{color}]", classes="dot"
                ),
                Label(f"{template.label}", classes="label"),
                id=f"template-{template.id}",
                classes="template-item",
            )
            widget.border_subtitle = str(index + 1)
            widget.can_focus = True
            container.compose_add_child(widget)
        return container

    def rebuild(self) -> None:
        self.templates = get_all_templates()
        container = self.query(".templates")
        if len(container) > 0:
            container[0].remove()
        container = Horizontal(classes="templates")
        container = self._create_templates_widgets(container)
        self.mount(container)
        if self.selected_template_id:
            self.call_after_refresh(
                lambda: self.query_one(f"#template-{self.selected_template_id}").focus()
            )

    # helper
    # -------------- Helper -------------- #

    def _notify_no_selected_template(self) -> None:
        self.app.notify(
            title="Error",
            message="No template selected",
            severity="error",
            timeout=3,
        )

    # ------------- Callback ------------- #

    def on_descendant_focus(self, event: events.DescendantFocus):
        id = event.widget.id
        if id:
            template_id = id.split("-")[1]
            self.selected_template_id = template_id

    def select_template(self, index: int) -> None:
        if index > len(self.templates):
            self.app.notify(
                title="Error",
                message=f"Template slot {index} is empty",
                severity="error",
                timeout=3,
            )
            return
        template = self.templates[index - 1]
        record_data = template.to_dict()
        record_data["date"] = self.page_parent.mode["date"]
        create_record(record_data)
        self.app.notify(
            title="Success",
            message=f"Created new record with {template.label}",
            severity="information",
            timeout=3,
        )
        self.page_parent.rebuild()

    def action_new_template(self) -> None:
        def check_result(result) -> None:
            if result:
                create_template(result)
                # try:
                # except Exception as e:
                #     self.app.notify(
                #         title="Error", message=f"{e}", severity="error", timeout=10
                #     )
                # else:
                self.app.notify(
                    title="Success",
                    message=f"Template created",
                    severity="information",
                    timeout=3,
                )
                self.rebuild()

        self.app.push_screen(
            InputModal("New Template", form=self.template_form.get_form()),
            callback=check_result,
        )

    def action_edit_template(self) -> None:
        if not self.selected_template_id:
            self._notify_no_selected_template()
            return

        # ----------------- - ---------------- #
        def check_result(result) -> None:
            if result:
                try:
                    update_template(self.selected_template_id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Template created",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        # ----------------- - ---------------- #
        self.app.push_screen(
            InputModal(
                "Edit Template",
                form=self.template_form.get_filled_form(self.selected_template_id),
            ),
            callback=check_result,
        )

    def action_delete_template(self) -> None:
        if not self.selected_template_id:
            self._notify_no_selected_template()
            return

        # ----------------- - ---------------- #
        def check_delete(result) -> None:
            if result:
                delete_template(self.selected_template_id)
                self.app.notify(
                    title="Success",
                    message=f"Template deleted",
                    severity="information",
                    timeout=3,
                )
                self.selected_template_id = None
                self.rebuild()

        # ----------------- - ---------------- #
        self.app.push_screen(
            ConfirmationModal("Are you sure you want to delete this template?"),
            check_delete,
        )

    def _swap_template(self, direction: str) -> None:
        if not self.selected_template_id:
            self._notify_no_selected_template()
            return
        if get_adjacent_template(self.selected_template_id, direction) == -1:
            self.app.notify(
                title="Error",
                message=f"No {direction} template",
                severity="error",
                timeout=3,
            )
            self.app.bell()
            return
        else:
            swap_template_order(self.selected_template_id, direction)
            self.rebuild()

    def action_swap_previous(self) -> None:
        self._swap_template("previous")

    def action_swap_next(self) -> None:
        self._swap_template("next")



from pydantic import BaseModel
from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label, Static


class BarchartData(BaseModel):
    amounts: list[float]
    labels: list[str]


class Barchart(Static):
    DEFAULT_CSS = """
    Barchart {
        width: 1fr;
        height: auto;

        & > .data-container {
            layout: horizontal;
            height: auto;

            .labels-container {
                height: auto;
                width: 5;
                color: $text-muted;
                border-right: wide $accent-lighten-3;
                background: $surface;
                
                .label {
                    width: 1fr;
                }
            }

            .bars-container {
            background: transparent;
            height: auto;

                .bar-container {
                    height: 1;
                    width: 100%;
                    background: $surface;
                    
                    .bar {
                        background: $surface-lighten-2;
                        min-width: 1;
                        border-top: tall $surface-lighten-3;
                    }
                }
            }
        }

        .max-amount {
            color: $text-muted;
            text-align: right;
            width: 1fr;
        }
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data = BarchartData(amounts=[], labels=[])
        self.last_count = 0

    def on_mount(self) -> None:
        self.rebuild()

    def set_data(self, data: BarchartData) -> None:
        self.data = data
        self.rebuild()

    def rebuild(self):
        if len(self.data.amounts) == 0:
            self.styles.display = "none"
            return
        else:
            self.styles.display = "block"

        max_amount = max(self.data.amounts)
        self.query_one(".max-amount").update(f"{max_amount}")

        # If count changed, do full rebuild
        if len(self.data.amounts) != self.last_count:
            # Clear existing bars and labels
            bars_container = self.query(".bars-container")
            if bars_container:
                bars_container[0].remove()

            labels_container = self.query(".labels-container")
            if labels_container:
                labels_container[0].remove()

            bars_container = Container(classes="bars-container")
            labels_container = Container(classes="labels-container")

            for i in range(len(self.data.amounts)):
                amount = self.data.amounts[i]
                label = self.data.labels[i]
                percentage = (amount / max_amount * 100) if max_amount > 0 else 0
                # build bar
                bar_container = Container(classes="bar-container")
                bar = Static(" ", classes="bar")
                bar.styles.width = f"{percentage}%"
                bar_container.compose_add_child(bar)
                bars_container.compose_add_child(bar_container)
                # build label
                label_widget = Label(label, classes="label")
                labels_container.compose_add_child(label_widget)

            data_container = self.query_one(".data-container")
            data_container.mount(labels_container)
            data_container.mount(bars_container)

            self.last_count = len(self.data.amounts)

        else:
            # Just update existing widgets
            bars = self.query(".bar")
            labels = self.query(".label")

            for i in range(len(self.data.amounts)):
                amount = self.data.amounts[i]
                label = self.data.labels[i]
                percentage = (amount / max_amount * 100) if max_amount > 0 else 0

                bars[i].styles.width = f"{percentage}%"
                labels[i].update(label)

    def compose(self) -> ComposeResult:
        # Create containers
        yield Label(f"Loading...", classes="max-amount")
        yield Container(classes="data-container")



from textual.app import ComposeResult
from textual.widgets import Static

from textual.widgets import Button
from bagels.config import CONFIG


class IncomeMode(Static):

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="incomemode-container", classes="module-container"
        )
        super().__setattr__("border_title", "View and add")
        super().__setattr__("border_subtitle", CONFIG.hotkeys.home.toggle_income_mode)
        self.page_parent = parent

    def on_mount(self) -> None:
        self.rebuild()

    # region Builder
    # -------------- Builder ------------- #

    def rebuild(self) -> None:
        expense_label: Button = self.query_one("#expense-label")
        income_label: Button = self.query_one("#income-label")
        current_is_income = self.page_parent.mode["isIncome"]
        expense_label.classes = "selected" if not current_is_income else ""
        income_label.classes = "selected" if current_is_income else ""

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.page_parent.action_toggle_income_mode()

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        yield Button("Expense", id="expense-label")
        yield Button("Income", id="income-label")



from typing import Any, Mapping, NamedTuple, Protocol, runtime_checkable

from textual.errors import NoWidget
from textual.geometry import Offset
from textual.screen import Screen
from textual.widget import Widget


@runtime_checkable
class Jumpable(Protocol):
    """A widget which we can jump focus to."""

    jump_key: str


class JumpInfo(NamedTuple):
    """Information returned by the jumper for each jump target."""

    key: str
    """The key which should trigger the jump."""

    widget: str | Widget
    """Either the ID or a direct reference to the widget."""


class Jumper:
    """An Amp-like jumping mechanism for quick spatial navigation"""

    def __init__(self, ids_to_keys: Mapping[str, str], screen: Screen[Any]) -> None:
        self.ids_to_keys = ids_to_keys
        self.keys_to_ids = {v: k for k, v in ids_to_keys.items()}
        self.screen = screen

    def get_overlays(self) -> dict[Offset, JumpInfo]:
        """Return a dictionary of all the jump targets"""
        screen = self.screen
        children: list[Widget] = screen.walk_children(Widget)
        overlays: dict[Offset, JumpInfo] = {}
        ids_to_keys = self.ids_to_keys
        for child in children:
            try:
                widget_offset = screen.get_offset(child)
            except NoWidget:
                # The widget might not be visible in the layout
                # due to it being hidden in some modes.
                continue

            if child.id and child.id in ids_to_keys:
                overlays[widget_offset] = JumpInfo(
                    ids_to_keys[child.id],
                    child.id,
                )
            elif isinstance(child, Jumpable):
                overlays[widget_offset] = JumpInfo(
                    child.jump_key,
                    child,
                )

        return overlays



from typing import TYPE_CHECKING

from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center
from textual.geometry import Offset
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Label

from bagels.components.jumper import JumpInfo
from bagels.config import CONFIG

if TYPE_CHECKING:
    from bagels.components.jumper import Jumper


class JumpOverlay(ModalScreen[str | Widget | None]):
    """Overlay showing the jump targets.
    Dismissed with the ID of the widget the jump was requested for on closing,
    or a reference to the widget. Is dismissed with None if the user dismissed
    the overlay without making a selection."""

    DEFAULT_CSS = """\
    JumpOverlay {
        background: black 25%;
    }
    """

    BINDINGS = [
        Binding("escape", "dismiss_overlay", "Dismiss", show=False),
    ]

    def __init__(
        self,
        jumper: "Jumper",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes)
        self.jumper: Jumper = jumper
        self.keys_to_widgets: dict[str, Widget | str] = {}
        self._resize_counter = 0

    def on_key(self, key_event: events.Key) -> None:
        # We need to stop the bubbling of these keys, because if they
        # arrive at the parent after the overlay is closed, then the parent
        # will handle the key event, resulting in the focus being shifted
        # again (unexpectedly) after the jump target was focused.
        if key_event.key in [
            "tab",
            "shift+tab",
            CONFIG.hotkeys.new,
            CONFIG.hotkeys.edit,
            CONFIG.hotkeys.delete,
        ]:
            key_event.stop()
            key_event.prevent_default()

        if self.is_active:
            # If they press a key corresponding to a jump target,
            # then we jump to it.
            target = self.keys_to_widgets.get(key_event.key)
            if target is not None:
                self.dismiss(target)
                return

    def action_dismiss_overlay(self) -> None:
        self.dismiss(None)

    async def on_resize(self) -> None:
        self._resize_counter += 1
        if self._resize_counter == 1:
            return
        await self.recompose()

    def _sync(self) -> None:
        self.overlays = self.jumper.get_overlays()
        self.keys_to_widgets = {v.key: v.widget for v in self.overlays.values()}

    def compose(self) -> ComposeResult:
        self._sync()
        for offset, jump_info in self.overlays.items():
            key, _widget = jump_info
            label = Label(key, classes="textual-jump-label")
            label.styles.offset = offset
            yield label
        with Center(id="textual-jump-info"):
            yield Label("Press a key to jump")
        with Center(id="textual-jump-dismiss"):
            yield Label("[b]ESC[/] to dismiss")



from sqlalchemy import select
from sqlalchemy.orm import joinedload, sessionmaker
from bagels.models.database.app import db_engine
from bagels.models.record_template import RecordTemplate

Session = sessionmaker(bind=db_engine)


# region c
def create_template(data):
    session = Session()
    try:
        new_template = RecordTemplate(**data)
        session.add(new_template)
        session.commit()
        session.refresh(new_template)
        session.expunge(new_template)
        return new_template
    finally:
        session.close()


# region r
def get_all_templates():
    session = Session()
    try:
        stmt = (
            select(RecordTemplate)
            .options(
                joinedload(RecordTemplate.category),
                joinedload(RecordTemplate.account),
            )
            .order_by(RecordTemplate.order)
        )
        return session.scalars(stmt).all()
    finally:
        session.close()


def get_template_by_id(recordtemplate_id):
    session = Session()
    try:
        stmt = (
            select(RecordTemplate)
            .options(
                joinedload(RecordTemplate.category),
                joinedload(RecordTemplate.account),
            )
        )
        return session.get(RecordTemplate, recordtemplate_id, options=[
            joinedload(RecordTemplate.category),
            joinedload(RecordTemplate.account),
        ])
    finally:
        session.close()


def get_adjacent_template(recordtemplate_id, direction):
    session = Session()
    try:
        recordtemplate = session.get(RecordTemplate, recordtemplate_id)
        if not recordtemplate:
            return -1

        current_order = recordtemplate.order
        if direction == "next":
            stmt = (
                select(RecordTemplate)
                .filter(RecordTemplate.order == current_order + 1)
            )
        else:  # direction == "previous"
            stmt = (
                select(RecordTemplate)
                .filter(RecordTemplate.order == current_order - 1)
            )

        adjacent_template = session.scalars(stmt).first()
        if adjacent_template:
            return adjacent_template.id
        return -1
    finally:
        session.close()


# region u
def update_template(recordtemplate_id, data):
    session = Session()
    try:
        recordtemplate = session.get(RecordTemplate, recordtemplate_id)
        if recordtemplate:
            for key, value in data.items():
                setattr(recordtemplate, key, value)
            session.commit()
            session.refresh(recordtemplate)
            session.expunge(recordtemplate)
        return recordtemplate
    finally:
        session.close()


def swap_template_order(recordtemplate_id, direction="next"):
    session = Session()
    try:
        recordtemplate = session.get(RecordTemplate, recordtemplate_id)

        if recordtemplate:
            current_order = recordtemplate.order
            if direction == "next":
                stmt = (
                    select(RecordTemplate)
                    .filter(RecordTemplate.order == current_order + 1)
                )
            else:  # direction == "previous"
                stmt = (
                    select(RecordTemplate)
                    .filter(RecordTemplate.order == current_order - 1)
                )

            swap_template = session.scalars(stmt).first()
            if swap_template:
                # Use negative order values to avoid unique constraint violations
                recordtemplate.order = -current_order
                session.flush()
                
                swap_template.order = -swap_template.order
                session.flush()
                
                recordtemplate.order = -swap_template.order
                swap_template.order = current_order
                session.commit()
                session.refresh(recordtemplate)
                session.expunge(recordtemplate)
        return recordtemplate
    finally:
        session.close()


# region d
def delete_template(recordtemplate_id):
    session = Session()
    try:
        recordtemplate = session.get(RecordTemplate, recordtemplate_id)
        if recordtemplate:
            # Get all templates with higher order
            stmt = select(RecordTemplate).filter(RecordTemplate.order > recordtemplate.order).order_by(RecordTemplate.order)
            templates = session.scalars(stmt).all()
            
            # First update all orders to temporary negative values
            for i, template in enumerate(templates):
                template.order = -(i + 1)
            session.flush()
            
            # Delete the target template
            session.delete(recordtemplate)
            session.flush()
            
            # Then update to final values
            for i, template in enumerate(templates):
                template.order = recordtemplate.order + i
            
            session.commit()
            return True
        return False
    finally:
        session.close()



from typing import Any, List, Literal
from pydantic import BaseModel, Field
from rich.console import RenderableType


class Option(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    text: str | None = None  # if not provided, the value will be used
    value: Any
    prefix: RenderableType | None = None
    postfix: RenderableType | None = None


class Options(BaseModel):
    items: List[Option] = Field(default_factory=list)

    def __len__(self):
        return len(self.items)


class FormField(BaseModel):
    placeholder: str | None = None
    title: str | None = None
    key: str
    type: Literal[
        "string",
        "number",
        "integer",
        "boolean",
        "autocomplete",
        "dateAutoDay",
        "hidden",
    ]
    is_required: bool = False
    min: float | int | None = None
    max: float | int | None = None
    labels: List[str] | None = None  # for type "boolean"
    options: Options | None = None  # for type "autocomplete"
    default_value: Any = None
    default_value_text: str | None = None
    create_action: bool | None = None  # for type "autocomplete"


class Form(BaseModel):
    fields: List[FormField] = Field(default_factory=list)

    def __len__(self):
        return len(self.fields)
